from enum import Enum
from typing import Any, Dict


class DataType(Enum):
    TYPE_BOOL = 1
    TYPE_UINT8 = 2
    TYPE_UINT16 = 3
    TYPE_UINT32 = 4
    TYPE_UINT64 = 5
    TYPE_INT8 = 6
    TYPE_INT16 = 7
    TYPE_INT32 = 8
    TYPE_INT64 = 9
    TYPE_FP16 = 10
    TYPE_FP32 = 11
    TYPE_FP64 = 12
    TYPE_STRING = 13


class TextGenerationInput:
    prompt = ""
    max_new_tokens = 100
    top_k = 1
    temperature = 0.8
    random_seed = 0
    stop_words: Any = ""
    extra_params: Dict[str, str] = {}



# pylint: disable=no-name-in-module
import os
import sys
from importlib.metadata import PackageNotFoundError, version

from instill.helpers.protobufs.ray_pb2 import (
    InferTensor,
    ModelInferRequest,
    ModelInferResponse,
    ModelMetadataRequest,
    ModelMetadataResponse,
    ModelReadyRequest,
    ModelReadyResponse,
)
from instill.utils.logger import Logger

Logger.initialize()

sys.path.append(os.path.join(os.path.dirname(__file__), "protogen"))

try:
    __version__ = version("instill-sdk")
except PackageNotFoundError:
    __version__ = "(local)"

del PackageNotFoundError
del version



from abc import ABC, abstractmethod


class Client(ABC):
    """Base interface class for creating mgmt/pipeline/connector/model clients.

    Args:
        ABC (abc.ABCMeta): std abstract class
    """

    @property
    @abstractmethod
    def hosts(self):
        pass

    @hosts.setter
    @abstractmethod
    def hosts(self):
        pass

    @property
    @abstractmethod
    def instance(self):
        pass

    @instance.setter
    @abstractmethod
    def instance(self):
        pass

    @property
    @abstractmethod
    def metadata(self):
        pass

    @metadata.setter
    @abstractmethod
    def metadata(self):
        pass

    @abstractmethod
    def liveness(self):
        raise NotImplementedError

    @abstractmethod
    def readiness(self):
        raise NotImplementedError

    @abstractmethod
    def is_serving(self):
        raise NotImplementedError



# pylint: disable=no-member,wrong-import-position,no-name-in-module
from instill.clients import InstillClient
from instill.resources.connector import Connector


class NumbersConnector(Connector):
    def __init__(
        self,
        client: InstillClient,
        name: str,
        capture_token: str,
        asset_type: str,
        metadata_texts: bool,
        metadata_structured_data: bool,
        metadata_metadata: bool,
    ) -> None:
        definition = "connector-definitions/blockchain-numbers"
        configuration = {
            "capture_token": capture_token,
            "asset_type": asset_type,
            "metadata_texts": metadata_texts,
            "metadata_structured_data": metadata_structured_data,
            "metadata_metadata": metadata_metadata,
        }
        super().__init__(client, name, definition, configuration)



# pylint: disable=no-member,wrong-import-position,no-name-in-module
from instill.clients import InstillClient
from instill.resources.connector import Connector


class PineconeConnector(Connector):
    def __init__(
        self,
        client: InstillClient,
        name: str,
        api_key: str,
        server_url: str,
    ) -> None:
        definition = "connector-definitions/data-pinecone"
        configuration = {"url": server_url, "api_key": api_key}
        super().__init__(client, name, definition, configuration)



import logging
import os

home_dir = os.path.expanduser("~")


class Logger:
    Initialized = False

    @staticmethod
    def initialize(
        runlog_path=f"{home_dir}/.config/instill/sdk/python/run.log",
        errorlog_path=f"{home_dir}/.config/instill/sdk/python/error.log",
    ):
        os.makedirs(os.path.dirname(runlog_path), exist_ok=True)
        os.makedirs(os.path.dirname(errorlog_path), exist_ok=True)

        logger = logging.getLogger()

        # remove handlers added by installed package
        for handler in logger.handlers[:]:
            logger.removeHandler(handler)

        logger.setLevel(logging.INFO)
        formatter = logging.Formatter(
            "%(asctime)s.%(msecs)03d %(levelname)-8s %(message)s"
        )
        runlog_handler = logging.FileHandler(runlog_path)
        runlog_handler.setLevel(logging.INFO)
        errorlog_handler = logging.FileHandler(errorlog_path)
        errorlog_handler.setLevel(logging.ERROR)
        console_handler = logging.StreamHandler()
        console_handler.setLevel(logging.INFO)

        runlog_handler.setFormatter(formatter)
        errorlog_handler.setFormatter(formatter)
        console_handler.setFormatter(formatter)

        logger.addHandler(runlog_handler)
        logger.addHandler(errorlog_handler)
        logger.addHandler(console_handler)
        Logger.Initialized = True

    @staticmethod
    def exception(message):
        assert Logger.Initialized, "Logger has not been initialized"
        logging.exception(message)

    @staticmethod
    def log(level, message):
        assert Logger.Initialized, "Logger has not been initialized"
        logging.log(level, message)

    @staticmethod
    def d(message):
        Logger.log(logging.DEBUG, message)

    @staticmethod
    def i(message):
        Logger.log(logging.INFO, message)

    @staticmethod
    def w(message):
        Logger.log(logging.WARNING, message)

    @staticmethod
    def e(message):
        Logger.log(logging.ERROR, message)



import json
import struct
from typing import List

import numpy as np

from instill.helpers.const import TextGenerationInput


def serialize_byte_tensor(input_tensor):
    """
    Serializes a bytes tensor into a flat numpy array of length prepended
    bytes. The numpy array should use dtype of np.object_. For np.bytes_,
    numpy will remove trailing zeros at the end of byte sequence and because
    of this it should be avoided.
    Parameters
    ----------
    input_tensor : np.array
        The bytes tensor to serialize.
    Returns
    -------
    serialized_bytes_tensor : np.array
        The 1-D numpy array of type uint8 containing the serialized bytes in 'C' order.
    Raises
    ------
    InferenceServerException
        If unable to serialize the given tensor.
    """

    if input_tensor.size == 0:
        return ()

    # If the input is a tensor of string/bytes objects, then must flatten those
    # into a 1-dimensional array containing the 4-byte byte size followed by the
    # actual element bytes. All elements are concatenated together in "C" order.
    if (input_tensor.dtype == np.object_) or (input_tensor.dtype.type == np.bytes_):
        flattened_ls: list = []
        for obj in np.nditer(input_tensor, flags=["refs_ok"], order="C"):
            # If directly passing bytes to BYTES type,
            # don't convert it to str as Python will encode the
            # bytes which may distort the meaning
            assert isinstance(obj, np.ndarray)
            if input_tensor.dtype == np.object_:
                if isinstance(obj.item(), bytes):
                    s = obj.item()
                else:
                    s = str(obj.item()).encode("utf-8")
            else:
                s = obj.item()
            flattened_ls.append(struct.pack("<I", len(s)))
            flattened_ls.append(s)
        flattened = b"".join(flattened_ls)
        return flattened
    return None


def deserialize_bytes_tensor(encoded_tensor):
    """
    Deserializes an encoded bytes tensor into an
    numpy array of dtype of python objects

    Parameters
    ----------
    encoded_tensor : bytes
        The encoded bytes tensor where each element
        has its length in first 4 bytes followed by
        the content
    Returns
    -------
    string_tensor : np.array
        The 1-D numpy array of type object containing the
        deserialized bytes in 'C' order.

    """
    strs = []
    offset = 0
    val_buf = encoded_tensor
    while offset < len(val_buf):
        l = struct.unpack_from("<I", val_buf, offset)[0]
        offset += 4
        sb = struct.unpack_from(f"<{l}s", val_buf, offset)[0]
        offset += l
        strs.append(sb)
    return np.array(strs, dtype=bytes)


class StandardTaskIO:
    @staticmethod
    def parse_task_text_generation_input(request) -> TextGenerationInput:
        text_generation_input = TextGenerationInput()

        for i, b_input_tensor in zip(request.inputs, request.raw_input_contents):
            input_name = i.name

            if input_name == "prompt":
                input_tensor = deserialize_bytes_tensor(b_input_tensor)
                text_generation_input.prompt = str(input_tensor[0].decode("utf-8"))
                print(
                    f"[DEBUG] input `prompt` type\
                        ({type(text_generation_input.prompt)}): {text_generation_input.prompt}"
                )

            if input_name == "max_new_tokens":
                text_generation_inputmax_new_tokens = int.from_bytes(
                    b_input_tensor, "little"
                )
                print(
                    f"[DEBUG] input `max_new_tokens` type\
                        ({type(text_generation_inputmax_new_tokens)}): {text_generation_inputmax_new_tokens}"
                )

            if input_name == "top_k":
                text_generation_input.top_k = int.from_bytes(b_input_tensor, "little")
                print(
                    f"[DEBUG] input `top_k` type\
                        ({type(text_generation_input.top_k)}): {text_generation_input.top_k}"
                )

            if input_name == "temperature":
                text_generation_input.temperature = struct.unpack("f", b_input_tensor)[
                    0
                ]
                print(
                    f"[DEBUG] input `temperature` type\
                        ({type(text_generation_input.temperature)}): {text_generation_input.temperature}"
                )
                text_generation_input.temperature = round(
                    text_generation_input.temperature, 2
                )

            if input_name == "random_seed":
                text_generation_input.random_seed = int.from_bytes(
                    b_input_tensor, "little"
                )
                print(
                    f"[DEBUG] input `random_seed` type\
                        ({type(text_generation_input.random_seed)}): {text_generation_input.random_seed}"
                )

            if input_name == "stop_words":
                input_tensor = deserialize_bytes_tensor(b_input_tensor)
                text_generation_input.stop_words = input_tensor[0]
                print(
                    f"[DEBUG] input `stop_words` type\
                        ({type(text_generation_input.stop_words)}): {text_generation_input.stop_words}"
                )
                if len(text_generation_input.stop_words) == 0:
                    text_generation_input.stop_words = None
                elif text_generation_input.stop_words.shape[0] > 1:
                    # TODO: Check wether shoule we decode this words
                    text_generation_input.stop_words = list(
                        text_generation_input.stop_words
                    )
                else:
                    text_generation_input.stop_words = [
                        str(text_generation_input.stop_words[0])
                    ]
                print(
                    f"[DEBUG] parsed input `stop_words` type\
                        ({type(text_generation_input.stop_words)}): {text_generation_input.stop_words}"
                )

            if input_name == "extra_params":
                input_tensor = deserialize_bytes_tensor(b_input_tensor)
                extra_params_str = str(input_tensor[0].decode("utf-8"))
                print(
                    f"[DEBUG] input `extra_params` type\
                        ({type(extra_params_str)}): {extra_params_str}"
                )

                try:
                    text_generation_input.extra_params = json.loads(extra_params_str)
                except json.decoder.JSONDecodeError:
                    print("[DEBUG] WARNING `extra_params` parsing faield!")
                    continue

        return text_generation_input

    @staticmethod
    def parse_task_text_generation_output(sequences: list):
        text_outputs = [seq["generated_text"].encode("utf-8") for seq in sequences]

        return serialize_byte_tensor(np.asarray(text_outputs))


class RawIO:
    @staticmethod
    def parse_byte_tensor(byte_tensor) -> List[str]:
        input_tensors = deserialize_bytes_tensor(byte_tensor)
        outs = [str(tensor.decode("utf-8")) for tensor in input_tensors]

        return outs

    @staticmethod
    def parse_unsigned_int_tensor(int_tensor) -> int:
        return int.from_bytes(int_tensor, "little")

    @staticmethod
    def parse_signed_int_tensor(int_tensor) -> int:
        return int.from_bytes(int_tensor, "little", signed=True)

    @staticmethod
    def parse_float_tensor(float_tensor) -> float:
        return struct.unpack("f", float_tensor)[0]

    @staticmethod
    def parse_boolean_tensor(bool_tensor) -> bool:
        return struct.unpack("?", bool_tensor)[0]



# pylint: disable=no-member,wrong-import-position
from collections import defaultdict
from typing import Iterable, Tuple, Union

import grpc
from google.longrunning import operations_pb2
from google.protobuf import field_mask_pb2

import instill.protogen.common.healthcheck.v1alpha.healthcheck_pb2 as healthcheck

# pipeline
import instill.protogen.vdp.pipeline.v1alpha.connector_pb2 as connector_interface
import instill.protogen.vdp.pipeline.v1alpha.operator_definition_pb2 as operator_interface
import instill.protogen.vdp.pipeline.v1alpha.pipeline_pb2 as pipeline_interface
import instill.protogen.vdp.pipeline.v1alpha.pipeline_public_service_pb2_grpc as pipeline_service
from instill.clients import constant

# common
from instill.clients.base import Client
from instill.configuration import global_config
from instill.utils.error_handler import grpc_handler

# from instill.utils.logger import Logger


class PipelineClient(Client):
    def __init__(self, namespace: str) -> None:
        self.hosts: defaultdict = defaultdict(dict)
        self.namespace: str = namespace
        if constant.DEFAULT_INSTANCE in global_config.hosts:
            self.instance = constant.DEFAULT_INSTANCE
        elif len(global_config.hosts) == 0:
            self.instance = ""
        else:
            self.instance = list(global_config.hosts.keys())[0]

        if global_config.hosts is not None:
            for instance, config in global_config.hosts.items():
                if not config.secure:
                    channel = grpc.insecure_channel(config.url)
                    self.hosts[instance]["metadata"] = (
                        (
                            "authorization",
                            f"Bearer {config.token}",
                        ),
                    )
                else:
                    ssl_creds = grpc.ssl_channel_credentials()
                    call_creds = grpc.access_token_call_credentials(config.token)
                    creds = grpc.composite_channel_credentials(ssl_creds, call_creds)
                    channel = grpc.secure_channel(
                        target=config.url,
                        credentials=creds,
                    )
                    self.hosts[instance]["metadata"] = ""
                self.hosts[instance]["token"] = config.token
                self.hosts[instance]["channel"] = channel
                self.hosts[instance][
                    "client"
                ] = pipeline_service.PipelinePublicServiceStub(channel)

    @property
    def hosts(self):
        return self._hosts

    @hosts.setter
    def hosts(self, hosts: str):
        self._hosts = hosts

    @property
    def instance(self):
        return self._instance

    @instance.setter
    def instance(self, instance: str):
        self._instance = instance

    @property
    def metadata(self):
        return self._metadata

    @metadata.setter
    def metadata(self, metadata: str):
        self._metadata = metadata

    def liveness(self) -> healthcheck.HealthCheckResponse.ServingStatus:
        resp: pipeline_interface.LivenessResponse = self.hosts[self.instance][
            "client"
        ].Liveness(request=pipeline_interface.LivenessRequest())
        return resp.health_check_response.status

    def readiness(self) -> healthcheck.HealthCheckResponse.ServingStatus:
        resp: pipeline_interface.ReadinessResponse = self.hosts[self.instance][
            "client"
        ].Readiness(request=pipeline_interface.ReadinessRequest())
        return resp.health_check_response.status

    def is_serving(self) -> bool:
        try:
            return (
                self.readiness()
                == healthcheck.HealthCheckResponse.SERVING_STATUS_SERVING
            )
        except Exception:
            return False

    @grpc_handler
    def list_operator_definitions(
        self,
        filer_str: str = "",
        next_page_token: str = "",
        total_size: int = 100,
    ) -> Tuple[Iterable, str, int]:
        resp: operator_interface.ListOperatorDefinitionsResponse = self.hosts[
            self.instance
        ]["client"].ListUserPipelines(
            request=operator_interface.ListOperatorDefinitionsRequest(
                filter=filer_str,
                page_size=total_size,
                page_token=next_page_token,
                view=operator_interface.ListOperatorDefinitionsRequest.VIEW_FULL,
            ),
            metadata=self.hosts[self.instance]["metadata"],
        )

        return resp.operator_definitions, resp.next_page_token, resp.total_size

    @grpc_handler
    def get_operator_definition(
        self, name: str
    ) -> operator_interface.OperatorDefinition:
        resp: operator_interface.GetOperatorDefinitionResponse = self.hosts[
            self.instance
        ]["client"].GetOperatorDefinition(
            request=operator_interface.GetOperatorDefinitionRequest(
                name=f"operator-definitions//{name}",
                view=operator_interface.GetOperatorDefinitionRequest.VIEW_FULL,
            ),
            metadata=self.hosts[self.instance]["metadata"],
        )
        return resp.operator_definition

    @grpc_handler
    def create_pipeline(
        self,
        name: str,
        recipe: pipeline_interface.Recipe,
    ) -> pipeline_interface.Pipeline:
        pipeline = pipeline_interface.Pipeline(
            id=name,
            recipe=recipe,
        )
        resp: pipeline_interface.CreateUserPipelineResponse = self.hosts[self.instance][
            "client"
        ].CreateUserPipeline(
            request=pipeline_interface.CreateUserPipelineRequest(
                pipeline=pipeline, parent=self.namespace
            ),
            metadata=self.hosts[self.instance]["metadata"],
        )
        return resp.pipeline

    @grpc_handler
    def get_pipeline(self, name: str) -> pipeline_interface.Pipeline:
        resp: pipeline_interface.GetUserPipelineResponse = self.hosts[self.instance][
            "client"
        ].GetUserPipeline(
            request=pipeline_interface.GetUserPipelineRequest(
                name=f"{self.namespace}/pipelines/{name}"
            ),
            metadata=self.hosts[self.instance]["metadata"],
        )
        return resp.pipeline

    @grpc_handler
    def lookup_pipeline(self, pipeline_uid: str) -> pipeline_interface.Pipeline:
        resp: pipeline_interface.LookUpPipelineResponse = self.hosts[self.instance][
            "client"
        ].LookUpPipeline(
            request=pipeline_interface.LookUpPipelineRequest(
                permalink=f"pipelines/{pipeline_uid}",
                view=pipeline_interface.LookUpPipelineRequest.VIEW_FULL,
            ),
            metadata=self.hosts[self.instance]["metadata"],
        )
        return resp.pipeline

    @grpc_handler
    def rename_pipeline(self, name: str, new_name: str) -> pipeline_interface.Pipeline:
        resp: pipeline_interface.RenameUserPipelineResponse = self.hosts[self.instance][
            "client"
        ].RenameUserPipeline(
            request=pipeline_interface.RenameUserPipelineRequest(
                name=f"{self.namespace}/pipelines/{name}",
                new_pipeline_id=new_name,
            ),
            metadata=self.hosts[self.instance]["metadata"],
        )
        return resp.pipeline

    @grpc_handler
    def update_pipeline(
        self, pipeline: pipeline_interface.Pipeline, mask: field_mask_pb2.FieldMask
    ) -> pipeline_interface.Pipeline:
        resp: pipeline_interface.UpdateUserPipelineResponse = self.hosts[self.instance][
            "client"
        ].UpdateUserPipeline(
            request=pipeline_interface.UpdateUserPipelineRequest(
                pipeline=pipeline,
                update_mask=mask,
            ),
            metadata=self.hosts[self.instance]["metadata"],
        )
        return resp.pipeline

    @grpc_handler
    def validate_pipeline(self, name: str) -> pipeline_interface.Pipeline:
        resp: pipeline_interface.ValidateUserPipelineResponse = self.hosts[
            self.instance
        ]["client"].ValidateUserPipeline(
            request=pipeline_interface.ValidateUserPipelineRequest(
                name=f"{self.namespace}/pipelines/{name}"
            ),
            metadata=self.hosts[self.instance]["metadata"],
        )
        return resp.pipeline

    @grpc_handler
    def trigger_pipeline(
        self, name: str, inputs: list
    ) -> Tuple[Iterable, pipeline_interface.TriggerMetadata]:
        resp: pipeline_interface.TriggerUserPipelineResponse = self.hosts[
            self.instance
        ]["client"].TriggerUserPipeline(
            request=pipeline_interface.TriggerUserPipelineRequest(
                name=f"{self.namespace}/pipelines/{name}", inputs=inputs
            ),
            metadata=self.hosts[self.instance]["metadata"],
        )
        return resp.outputs, resp.metadata

    @grpc_handler
    def trigger_async_pipeline(
        self, name: str, inputs: list
    ) -> operations_pb2.Operation:
        resp: pipeline_interface.TriggerAsyncUserPipelineResponse = self.hosts[
            self.instance
        ]["client"].TriggerAsyncUserPipeline(
            request=pipeline_interface.TriggerAsyncUserPipelineRequest(
                name=f"{self.namespace}/pipelines/{name}", inputs=inputs
            ),
            metadata=self.hosts[self.instance]["metadata"],
        )
        return resp.operation

    @grpc_handler
    def delete_pipeline(self, name: str):
        self.hosts[self.instance]["client"].DeleteUserPipeline(
            request=pipeline_interface.DeleteUserPipelineRequest(
                name=f"{self.namespace}/pipelines/{name}"
            ),
            metadata=self.hosts[self.instance]["metadata"],
        )

    @grpc_handler
    def list_pipelines(
        self,
        filer_str: str = "",
        next_page_token: str = "",
        total_size: int = 100,
        show_deleted: bool = False,
        public=False,
    ) -> Tuple[Iterable, str, int]:
        resp: Union[
            pipeline_interface.ListPipelinesResponse,
            pipeline_interface.ListUserPipelinesResponse,
        ]
        if not public:
            resp = self.hosts[self.instance]["client"].ListUserPipelines(
                request=pipeline_interface.ListUserPipelinesRequest(
                    parent=self.namespace,
                    filter=filer_str,
                    page_size=total_size,
                    page_token=next_page_token,
                    show_deleted=show_deleted,
                    view=pipeline_interface.ListUserPipelinesRequest.VIEW_FULL,
                ),
                metadata=self.hosts[self.instance]["metadata"],
            )
        else:
            resp = self.hosts[self.instance]["client"].ListPipelines(
                request=pipeline_interface.ListPipelinesRequest(
                    filter=filer_str,
                    page_size=total_size,
                    page_token=next_page_token,
                    show_deleted=show_deleted,
                    view=pipeline_interface.ListPipelinesRequest.VIEW_FULL,
                ),
                metadata=self.hosts[self.instance]["metadata"],
            )

        return resp.pipelines, resp.next_page_token, resp.total_size

    @grpc_handler
    def get_operation(self, name: str) -> operations_pb2.Operation:
        resp: pipeline_interface.GetOperationResponse = self.hosts[self.instance][
            "client"
        ].GetOperation(
            request=pipeline_interface.GetOperationRequest(
                name=name,
            ),
            metadata=self.hosts[self.instance]["metadata"],
        )
        return resp.operation

    @grpc_handler
    def create_pipeline_release(
        self,
        version: str,
    ) -> pipeline_interface.PipelineRelease:
        """Create a release version of a pipeline

        Args:
            name (str): Must be a sematic version vX.Y.Z

        Returns:
            pipeline_interface.PipelineRelease: Released pipeline
        """
        pipeline_release = pipeline_interface.PipelineRelease(
            id=version,
        )
        resp: pipeline_interface.CreateUserPipelineReleaseResponse = self.hosts[
            self.instance
        ]["client"].CreateUserPipelineRelease(
            request=pipeline_interface.CreateUserPipelineReleaseRequest(
                release=pipeline_release, parent=self.namespace
            ),
            metadata=self.hosts[self.instance]["metadata"],
        )
        return resp.release

    @grpc_handler
    def get_pipeline_release(self, name: str) -> pipeline_interface.PipelineRelease:
        """Get a released pipeline

        Args:
            name (str): Must have the format of {name}/releases/*

        Returns:
            pipeline_interface.Pipeline: Released pipeline
        """
        resp: pipeline_interface.GetUserPipelineReleaseResponse = self.hosts[
            self.instance
        ]["client"].GetUserPipelineRelease(
            request=pipeline_interface.GetUserPipelineReleaseRequest(
                name=f"{self.namespace}/pipelines/{name}",
                view=pipeline_interface.GetUserPipelineReleaseRequest.VIEW_FULL,
            ),
            metadata=self.hosts[self.instance]["metadata"],
        )
        return resp.release

    @grpc_handler
    def rename_pipeline_release(
        self, name: str, new_version: str
    ) -> pipeline_interface.PipelineRelease:
        """Rename a released pipeline

        Args:
            name (str): Must have the format of {name}/releases/*
            new_version (str): New release version, must be a sematic version vX.Y.Z

        Returns:
            pipeline_interface.PipelineRelease: released pipeline
        """
        resp: pipeline_interface.RenameUserPipelineReleaseResponse = self.hosts[
            self.instance
        ]["client"].RenameUserPipelineRelease(
            request=pipeline_interface.RenameUserPipelineReleaseRequest(
                name=f"{self.namespace}/pipelines/{name}",
                new_pipeline_release_id=new_version,
            ),
            metadata=self.hosts[self.instance]["metadata"],
        )
        return resp.release

    @grpc_handler
    def update_pipeline_release(
        self,
        pipeline_release: pipeline_interface.PipelineRelease,
        mask: field_mask_pb2.FieldMask,
    ) -> pipeline_interface.PipelineRelease:
        resp: pipeline_interface.UpdateUserPipelineReleaseResponse = self.hosts[
            self.instance
        ]["client"].UpdateUserPipelineRelease(
            request=pipeline_interface.UpdateUserPipelineReleaseRequest(
                release=pipeline_release,
                update_mask=mask,
            ),
            metadata=self.hosts[self.instance]["metadata"],
        )
        return resp.release

    @grpc_handler
    def list_pipeline_releases(
        self,
        filer_str: str = "",
        next_page_token: str = "",
        total_size: int = 100,
        show_deleted: bool = False,
    ) -> Tuple[Iterable, str, int]:
        resp: pipeline_interface.ListUserPipelineReleasesResponse = self.hosts[
            self.instance
        ]["client"].ListUserPipelineReleases(
            request=pipeline_interface.ListUserPipelineReleasesRequest(
                parent=self.namespace,
                filter=filer_str,
                page_size=total_size,
                page_token=next_page_token,
                show_deleted=show_deleted,
                view=pipeline_interface.ListUserPipelineReleasesRequest.VIEW_FULL,
            ),
            metadata=self.hosts[self.instance]["metadata"],
        )

        return resp.releases, resp.next_page_token, resp.total_size

    @grpc_handler
    def delete_pipeline_release(self, name: str):
        self.hosts[self.instance]["client"].DeleteUserPipelineRelease(
            request=pipeline_interface.DeleteUserPipelineReleaseRequest(
                name=f"{self.namespace}/pipelines/{name}"
            ),
            metadata=self.hosts[self.instance]["metadata"],
        )

    @grpc_handler
    def set_default_pipeline_release(
        self, name: str
    ) -> pipeline_interface.PipelineRelease:
        resp: pipeline_interface.SetDefaultUserPipelineReleaseResponse = self.hosts[
            self.instance
        ]["client"].SetDefaultUserPipelineRelease(
            request=pipeline_interface.SetDefaultUserPipelineReleaseRequest(
                name=f"{self.namespace}/pipelines/{name}",
            ),
            metadata=self.hosts[self.instance]["metadata"],
        )
        return resp.release

    @grpc_handler
    def restore_pipeline_release(self, name: str) -> pipeline_interface.PipelineRelease:
        resp: pipeline_interface.RestoreUserPipelineReleaseResponse = self.hosts[
            self.instance
        ]["client"].RestoreUserPipelineRelease(
            request=pipeline_interface.RestoreUserPipelineReleaseRequest(
                name=f"{self.namespace}/pipelines/{name}",
            ),
            metadata=self.hosts[self.instance]["metadata"],
        )
        return resp.release

    @grpc_handler
    def watch_pipeline_release(self, name: str) -> pipeline_interface.State.ValueType:
        resp: pipeline_interface.WatchUserPipelineReleaseResponse = self.hosts[
            self.instance
        ]["client"].WatchUserPipelineRelease(
            request=pipeline_interface.WatchUserPipelineReleaseRequest(
                name=f"{self.namespace}/pipelines/{name}",
            ),
            metadata=self.hosts[self.instance]["metadata"],
        )
        return resp.state

    @grpc_handler
    def trigger_pipeline_release(
        self, name: str, inputs: list
    ) -> Tuple[Iterable, pipeline_interface.TriggerMetadata]:
        resp: pipeline_interface.TriggerUserPipelineReleaseResponse = self.hosts[
            self.instance
        ]["client"].TriggerUserPipelineReleas(
            request=pipeline_interface.TriggerUserPipelineReleaseRequest(
                name=f"{self.namespace}/pipelines/{name}", inputs=inputs
            ),
            metadata=self.hosts[self.instance]["metadata"],
        )
        return resp.outputs, resp.metadata

    @grpc_handler
    def trigger_async_pipeline_release(
        self, name: str, inputs: list
    ) -> operations_pb2.Operation:
        resp: pipeline_interface.TriggerAsyncUserPipelineReleaseResponse = self.hosts[
            self.instance
        ]["client"].TriggerAsyncUserPipelineRelease(
            request=pipeline_interface.TriggerAsyncUserPipelineReleaseRequest(
                name=f"{self.namespace}/pipelines/{name}", inputs=inputs
            ),
            metadata=self.hosts[self.instance]["metadata"],
        )
        return resp.operation

    @grpc_handler
    def create_connector(
        self,
        name: str,
        definition: str,
        configuration: dict,
    ) -> connector_interface.Connector:
        connector = connector_interface.Connector()
        connector.id = name
        connector.connector_definition_name = definition
        connector.configuration.update(configuration)
        resp = self.hosts[self.instance]["client"].CreateUserConnector(
            request=connector_interface.CreateUserConnectorRequest(
                connector=connector, parent=self.namespace
            ),
            metadata=self.hosts[self.instance]["metadata"],
        )

        return resp.connector

    @grpc_handler
    def get_connector(self, name: str) -> connector_interface.Connector:
        return (
            self.hosts[self.instance]["client"]
            .GetUserConnector(
                request=connector_interface.GetUserConnectorRequest(
                    name=f"{self.namespace}/connectors/{name}",
                    view=connector_interface.GetUserConnectorRequest.VIEW_FULL,
                ),
                metadata=self.hosts[self.instance]["metadata"],
            )
            .connector
        )

    @grpc_handler
    def test_connector(self, name: str) -> connector_interface.Connector.State:
        return (
            self.hosts[self.instance]["client"]
            .TestUserConnector(
                request=connector_interface.TestUserConnectorRequest(
                    name=f"{self.namespace}/connectors/{name}"
                ),
                metadata=self.hosts[self.instance]["metadata"],
            )
            .state
        )

    @grpc_handler
    def execute_connector(self, name: str, inputs: list) -> list:
        return (
            self.hosts[self.instance]["client"]
            .ExecuteUserConnector(
                request=connector_interface.ExecuteUserConnectorRequest(
                    name=f"{self.namespace}/connectors/{name}", inputs=inputs
                ),
                metadata=self.hosts[self.instance]["metadata"],
            )
            .outputs
        )

    @grpc_handler
    def watch_connector(self, name: str) -> connector_interface.Connector.State:
        return (
            self.hosts[self.instance]["client"]
            .WatchUserConnector(
                request=connector_interface.WatchUserConnectorRequest(
                    name=f"{self.namespace}/connectors/{name}"
                ),
                metadata=self.hosts[self.instance]["metadata"],
            )
            .state
        )

    @grpc_handler
    def delete_connector(self, name: str):
        self.hosts[self.instance]["client"].DeleteUserConnector(
            request=connector_interface.DeleteUserConnectorRequest(
                name=f"{self.namespace}/connectors/{name}"
            ),
            metadata=self.hosts[self.instance]["metadata"],
        )

    @grpc_handler
    def list_connectors(self, public=False) -> Tuple[list, str, int]:
        if not public:
            resp = self.hosts[self.instance]["client"].ListUserConnectors(
                request=connector_interface.ListUserConnectorsRequest(
                    parent=self.namespace
                ),
                metadata=self.hosts[self.instance]["metadata"],
            )
        else:
            resp = self.hosts[self.instance]["client"].ListConnectors(
                request=connector_interface.ListConnectorsRequest(),
                metadata=(self.hosts[self.instance]["metadata"],),
            )

        return resp.connectors, resp.next_page_token, resp.total_size



# pylint: disable=redefined-outer-name,unused-variable,expression-not-assigned,no-name-in-module

from collections import defaultdict

from instill.clients import MgmtClient, ModelClient, PipelineClient


def describe_client():
    def describe_instance():
        def when_not_set(expect):
            mgmt_client = MgmtClient()
            expect(mgmt_client.instance) == ""
            model_client = ModelClient(namespace="")
            expect(model_client.instance) == ""
            pipeline_client = PipelineClient(namespace="")
            expect(pipeline_client.instance) == ""

        def when_set_correct_type(expect):
            mgmt_client = MgmtClient()
            mgmt_client.instance = "staging"
            expect(mgmt_client.instance) == "staging"
            model_client = ModelClient(namespace="")
            model_client.instance = "staging"
            expect(model_client.instance) == "staging"
            pipeline_client = PipelineClient(namespace="")
            pipeline_client.instance = "staging"
            expect(pipeline_client.instance) == "staging"

    def describe_host():
        def when_not_set(expect):
            mgmt_client = MgmtClient()
            expect(mgmt_client.hosts) is None
            model_client = ModelClient(namespace="")
            expect(model_client.hosts) is None
            pipeline_client = PipelineClient(namespace="")
            expect(pipeline_client.hosts) is None

        def when_set_correct_type(expect):
            mgmt_client = MgmtClient()
            d = defaultdict(dict)  # type: ignore
            d["test_instance"] = dict({"url": "test_url"})
            mgmt_client.hosts = d
            expect(mgmt_client.hosts["test_instance"]["url"]) == "test_url"
            model_client = ModelClient(namespace="")
            model_client.hosts = d
            expect(model_client.hosts["test_instance"]["url"]) == "test_url"
            pipeline_client = PipelineClient(namespace="")
            pipeline_client.hosts = d
            expect(pipeline_client.hosts["test_instance"]["url"]) == "test_url"



# pylint: disable=no-member,wrong-import-position
import instill.protogen.vdp.pipeline.v1alpha.pipeline_pb2 as pipeline_pb


def create_start_operator(config: dict) -> pipeline_pb.Component:
    start_operator_component = pipeline_pb.Component()
    start_operator_component.id = "start"
    start_operator_component.definition_name = "operator-definitions/op-start"
    start_operator_component.configuration.update(config)

    return start_operator_component


def create_end_operator(config: dict) -> pipeline_pb.Component:
    end_operator_component = pipeline_pb.Component()
    end_operator_component.id = "end"
    end_operator_component.definition_name = "operator-definitions/op-end"
    end_operator_component.configuration.update(config)

    return end_operator_component



# pylint: disable=no-member,wrong-import-position,no-name-in-module
import instill.protogen.vdp.pipeline.v1alpha.connector_definition_pb2 as connector_definition_interface
import instill.protogen.vdp.pipeline.v1alpha.connector_pb2 as connector_interface
import instill.protogen.vdp.pipeline.v1alpha.pipeline_pb2 as pipeline_interface
from instill.clients import InstillClient
from instill.resources.resource import Resource


class Connector(Resource):
    def __init__(
        self,
        client: InstillClient,
        name: str,
        definition: str,
        configuration: dict,
    ) -> None:
        super().__init__()
        self.client = client
        connector = client.pipeline_service.get_connector(name=name, silent=True)
        if connector is None:
            connector = client.pipeline_service.create_connector(
                name=name,
                definition=definition,
                configuration=configuration,
            )
            if connector is None:
                raise BaseException("connector creation failed")

        self.resource = connector

    def __call__(self, task_inputs: list, mode="execute") -> list:
        if mode == "execute":
            return self.client.pipeline_service.execute_connector(
                self.resource.id, task_inputs
            )
        return self.client.pipeline_service.test_connector(
            self.resource.id, task_inputs
        )

    @property
    def client(self):
        return self._client

    @client.setter
    def client(self, client: InstillClient):
        self._client = client

    @property
    def resource(self):
        return self._resource

    @resource.setter
    def resource(self, resource: connector_interface.Connector):
        self._resource = resource

    def create_component(self, name: str, config: dict) -> pipeline_interface.Component:
        component = pipeline_interface.Component()
        component.id = name
        component.definition_name = self.get_definition().name
        component.resource_name = self.resource.name
        component.configuration.update(config)
        return component

    def get_definition(self) -> connector_definition_interface.ConnectorDefinition:
        return self.resource.connector_definition

    def get_state(self) -> connector_interface.Connector.State:
        return self.client.pipeline_service.watch_connector(self.resource.id)

    def test(self) -> connector_interface.Connector.State:
        return self.client.pipeline_service.test_connector(self.resource.id)

    def delete(self):
        if self.resource is not None:
            self.client.pipeline_service.delete_connector(self.resource.id)



# pylint: disable=no-member,wrong-import-position,no-name-in-module
from typing import Tuple, Union

from google.longrunning import operations_pb2

import instill.protogen.vdp.pipeline.v1alpha.pipeline_pb2 as pipeline_interface
from instill.clients import InstillClient
from instill.resources.resource import Resource


class Pipeline(Resource):
    def __init__(
        self,
        client: InstillClient,
        name: str,
        recipe: Union[pipeline_interface.Recipe, None] = None,
    ) -> None:
        super().__init__()
        self.client = client
        pipeline = client.pipeline_service.get_pipeline(name=name, silent=True)
        if pipeline is None:
            pipeline = client.pipeline_service.create_pipeline(name=name, recipe=recipe)
            if pipeline is None:
                raise BaseException("pipeline creation failed")

        self.resource = pipeline

    def __call__(
        self, task_inputs: list
    ) -> Tuple[list, pipeline_interface.TriggerMetadata]:
        return self.client.pipeline_service.trigger_pipeline(
            self.resource.id, task_inputs
        )

    @property
    def client(self):
        return self._client

    @client.setter
    def client(self, client: InstillClient):
        self._client = client

    @property
    def resource(self):
        return self._resource

    @resource.setter
    def resource(self, resource: pipeline_interface.Pipeline):
        self._resource = resource

    def _update(self):
        self.resource = self.client.pipeline_service.get_pipeline(name=self.resource.id)

    def get_operation(self, operation: operations_pb2.Operation):
        return self.client.pipeline_service.get_operation(operation.name)

    def trigger_async(self, task_inputs: list) -> operations_pb2.Operation:
        return self.client.pipeline_service.trigger_async_pipeline(
            self.resource.id, task_inputs
        )

    def get_recipe(self) -> str:
        return self.resource.recipe

    def validate_pipeline(self) -> pipeline_interface.Pipeline:
        return self.client.pipeline_service.validate_pipeline(name=self.resource.id)

    def delete(self):
        if self.resource is not None:
            self.client.pipeline_service.delete_pipeline(self.resource.id)



# pylint: disable=no-member,wrong-import-position,no-name-in-module
from instill.clients import InstillClient
from instill.resources.connector import Connector


class InstillModelConnector(Connector):
    def __init__(
        self,
        client: InstillClient,
        name: str,
        server_url: str,
    ) -> None:
        definition = "connector-definitions/ai-instill-model"
        configuration = {
            "api_token": client.pipeline_service.hosts[
                client.pipeline_service.instance
            ]["token"],
            "server_url": server_url,
        }
        super().__init__(client, name, definition, configuration)


class StabilityAIConnector(Connector):
    def __init__(
        self,
        client: InstillClient,
        name: str,
        api_key: str,
    ) -> None:
        definition = "connector-definitions/ai-stability-ai"
        configuration = {"api_key": api_key}
        super().__init__(client, name, definition, configuration)


class OpenAIConnector(Connector):
    def __init__(
        self,
        client: InstillClient,
        name: str,
        api_key: str,
    ) -> None:
        definition = "connector-definitions/ai-openai"
        configuration = {
            "api_key": api_key,
        }
        super().__init__(client, name, definition, configuration)



import argparse


class InstillRayModelConfig:
    def __init__(
        self,
        ray_actor_options: dict,
        ray_autoscaling_options: dict,
        max_concurrent_queries: int,
        og_model_path: str,
        ray_addr: str,
    ) -> None:
        self.ray_addr = ray_addr

        og_model_string_parts = og_model_path.split("/")

        self.ray_actor_options = ray_actor_options
        self.ray_autoscaling_options = ray_autoscaling_options
        self.max_concurrent_queries = max_concurrent_queries

        self.model_path = og_model_path
        self.application_name = og_model_string_parts[5]
        self.model_name = "_".join(og_model_string_parts[3].split("#")[:2])
        self.route_prefix = (
            f'/{self.model_name}/{og_model_string_parts[3].split("#")[3]}'
        )


def get_compose_ray_address(port: int):
    return f"ray://ray_server:{port}"


def entry(model_weight_name_or_folder: str):
    parser = argparse.ArgumentParser()

    ray_actor_options = {
        "num_cpus": 1,
    }
    max_concurrent_queries = 10
    ray_autoscaling_options = {
        "target_num_ongoing_requests_per_replica": 7,
        "initial_replicas": 1,
        "min_replicas": 0,
        "max_replicas": 5,
    }

    parser.add_argument(
        "--func", required=True, choices=["deploy", "undeploy"], help="deploy/undeploy"
    )
    parser.add_argument("--model", required=True, help="model path for the deployment")
    parser.add_argument(
        "--ray-addr", default=get_compose_ray_address(10001), help="ray head address"
    )
    parser.add_argument(
        "--ray-actor-options",
        default=ray_actor_options,
        help="custom actor options for the deployment",
    )
    parser.add_argument(
        "--ray-autoscaling-options",
        default=ray_autoscaling_options,
        help="custom autoscaling options for the deployment",
    )
    args = parser.parse_args()

    ray_addr = "ray://" + args.ray_addr.replace("9000", "10001")

    model_config = InstillRayModelConfig(
        ray_addr=ray_addr,
        ray_actor_options=args.ray_actor_options,
        ray_autoscaling_options=args.ray_autoscaling_options,
        max_concurrent_queries=max_concurrent_queries,
        og_model_path="/".join([args.model, model_weight_name_or_folder]),
    )

    return args.func, model_config



# pylint: disable=no-member,wrong-import-position
from collections import defaultdict

import grpc

import instill.protogen.common.healthcheck.v1alpha.healthcheck_pb2 as healthcheck

# mgmt
import instill.protogen.core.mgmt.v1alpha.metric_pb2 as metric_interface
import instill.protogen.core.mgmt.v1alpha.mgmt_pb2 as mgmt_interface
import instill.protogen.core.mgmt.v1alpha.mgmt_public_service_pb2_grpc as mgmt_service
from instill.clients import constant

# common
from instill.clients.base import Client
from instill.configuration import global_config
from instill.utils.error_handler import grpc_handler

# from instill.utils.logger import Logger


class MgmtClient(Client):
    def __init__(self) -> None:
        self.hosts: defaultdict = defaultdict(dict)
        if constant.DEFAULT_INSTANCE in global_config.hosts:
            self.instance = constant.DEFAULT_INSTANCE
        elif len(global_config.hosts) == 0:
            self.instance = ""
        else:
            self.instance = list(global_config.hosts.keys())[0]

        if global_config.hosts is not None:
            for instance, config in global_config.hosts.items():
                if not config.secure:
                    channel = grpc.insecure_channel(config.url)
                    self.hosts[instance]["metadata"] = (
                        (
                            "authorization",
                            f"Bearer {config.token}",
                        ),
                    )
                else:
                    ssl_creds = grpc.ssl_channel_credentials()
                    call_creds = grpc.access_token_call_credentials(config.token)
                    creds = grpc.composite_channel_credentials(ssl_creds, call_creds)
                    channel = grpc.secure_channel(
                        target=config.url,
                        credentials=creds,
                    )
                    self.hosts[instance]["metadata"] = ""
                self.hosts[instance]["token"] = config.token
                self.hosts[instance]["channel"] = channel
                self.hosts[instance]["client"] = mgmt_service.MgmtPublicServiceStub(
                    channel
                )

    @property
    def hosts(self):
        return self._hosts

    @hosts.setter
    def hosts(self, hosts: str):
        self._hosts = hosts

    @property
    def instance(self):
        return self._instance

    @instance.setter
    def instance(self, instance: str):
        self._instance = instance

    @property
    def metadata(self):
        return self._metadata

    @metadata.setter
    def metadata(self, metadata: str):
        self._metadata = metadata

    def liveness(self) -> healthcheck.HealthCheckResponse.ServingStatus:
        resp: mgmt_interface.LivenessResponse = self.hosts[self.instance][
            "client"
        ].Liveness(request=mgmt_interface.LivenessRequest())
        return resp.health_check_response.status

    def readiness(self) -> healthcheck.HealthCheckResponse.ServingStatus:
        resp: mgmt_interface.ReadinessResponse = self.hosts[self.instance][
            "client"
        ].Readiness(request=mgmt_interface.ReadinessRequest())
        return resp.health_check_response.status

    def is_serving(self) -> bool:
        try:
            return (
                self.readiness()
                == healthcheck.HealthCheckResponse.SERVING_STATUS_SERVING
            )
        except Exception:
            return False

    @grpc_handler
    def login(self, username="admin", password="password") -> str:
        resp: mgmt_interface.AuthLoginResponse = self.hosts[self.instance][
            "client"
        ].AuthLogin(
            request=mgmt_interface.AuthLoginRequest(
                username=username, password=password
            )
        )
        return resp.access_token

    @grpc_handler
    def get_token(self, name: str) -> mgmt_interface.ApiToken:
        resp: mgmt_interface.GetTokenResponse = self.hosts[self.instance][
            "client"
        ].GetToken(
            request=mgmt_interface.GetTokenRequest(name=name),
            metadata=self.hosts[self.instance]["metadata"],
        )
        return resp.token

    @grpc_handler
    def get_user(self) -> mgmt_interface.User:
        resp: mgmt_interface.GetUserResponse = self.hosts[self.instance][
            "client"
        ].GetUser(
            request=mgmt_interface.GetUserRequest(name="users/me"),
            metadata=self.hosts[self.instance]["metadata"],
        )
        return resp.user

    @grpc_handler
    def list_pipeline_trigger_records(
        self,
    ) -> metric_interface.ListPipelineTriggerRecordsResponse:
        return self.hosts[self.instance]["client"].ListPipelineTriggerRecords(
            request=metric_interface.ListPipelineTriggerChartRecordsRequest(),
            metadata=self.hosts[self.instance]["metadata"],
        )

    @grpc_handler
    def list_pipeline_trigger_table_records(
        self,
    ) -> metric_interface.ListPipelineTriggerTableRecordsRequest:
        return self.hosts[self.instance]["client"].ListPipelineTriggerRecords(
            request=metric_interface.ListPipelineTriggerTableRecordsResponse(),
            metadata=self.hosts[self.instance]["metadata"],
        )

    @grpc_handler
    def list_pipeline_trigger_chart_records(
        self,
    ) -> metric_interface.ListPipelineTriggerChartRecordsResponse:
        return self.hosts[self.instance]["client"].ListPipelineTriggerRecords(
            request=metric_interface.ListPipelineTriggerChartRecordsRequest(),
            metadata=self.hosts[self.instance]["metadata"],
        )

    @grpc_handler
    def list_connector_execute_records(
        self,
    ) -> metric_interface.ListConnectorExecuteRecordsResponse:
        return self.hosts[self.instance]["client"].ListPipelineTriggerRecords(
            request=metric_interface.ListConnectorExecuteRecordsRequest(),
            metadata=self.hosts[self.instance]["metadata"],
        )

    @grpc_handler
    def list_connector_execute_table_records(
        self,
    ) -> metric_interface.ListConnectorExecuteTableRecordsResponse:
        return self.hosts[self.instance]["client"].ListPipelineTriggerRecords(
            request=metric_interface.ListConnectorExecuteTableRecordsRequest(),
            metadata=self.hosts[self.instance]["metadata"],
        )

    @grpc_handler
    def list_connector_execute_chart_records(
        self,
    ) -> metric_interface.ListConnectorExecuteChartRecordsResponse:
        return self.hosts[self.instance]["client"].ListPipelineTriggerRecords(
            request=metric_interface.ListConnectorExecuteChartRecordsRequest(),
            metadata=self.hosts[self.instance]["metadata"],
        )



from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class ModelReadyRequest(_message.Message):
    __slots__ = ["name", "version"]
    NAME_FIELD_NUMBER: _ClassVar[int]
    VERSION_FIELD_NUMBER: _ClassVar[int]
    name: str
    version: str
    def __init__(self, name: _Optional[str] = ..., version: _Optional[str] = ...) -> None: ...

class ModelReadyResponse(_message.Message):
    __slots__ = ["ready"]
    READY_FIELD_NUMBER: _ClassVar[int]
    ready: bool
    def __init__(self, ready: bool = ...) -> None: ...

class ModelMetadataRequest(_message.Message):
    __slots__ = ["name", "version"]
    NAME_FIELD_NUMBER: _ClassVar[int]
    VERSION_FIELD_NUMBER: _ClassVar[int]
    name: str
    version: str
    def __init__(self, name: _Optional[str] = ..., version: _Optional[str] = ...) -> None: ...

class InferTensor(_message.Message):
    __slots__ = ["name", "datatype", "shape"]
    NAME_FIELD_NUMBER: _ClassVar[int]
    DATATYPE_FIELD_NUMBER: _ClassVar[int]
    SHAPE_FIELD_NUMBER: _ClassVar[int]
    name: str
    datatype: str
    shape: _containers.RepeatedScalarFieldContainer[int]
    def __init__(self, name: _Optional[str] = ..., datatype: _Optional[str] = ..., shape: _Optional[_Iterable[int]] = ...) -> None: ...

class ModelMetadataResponse(_message.Message):
    __slots__ = ["name", "versions", "framework", "inputs", "outputs"]
    class TensorMetadata(_message.Message):
        __slots__ = ["name", "datatype", "shape"]
        NAME_FIELD_NUMBER: _ClassVar[int]
        DATATYPE_FIELD_NUMBER: _ClassVar[int]
        SHAPE_FIELD_NUMBER: _ClassVar[int]
        name: str
        datatype: str
        shape: _containers.RepeatedScalarFieldContainer[int]
        def __init__(self, name: _Optional[str] = ..., datatype: _Optional[str] = ..., shape: _Optional[_Iterable[int]] = ...) -> None: ...
    NAME_FIELD_NUMBER: _ClassVar[int]
    VERSIONS_FIELD_NUMBER: _ClassVar[int]
    FRAMEWORK_FIELD_NUMBER: _ClassVar[int]
    INPUTS_FIELD_NUMBER: _ClassVar[int]
    OUTPUTS_FIELD_NUMBER: _ClassVar[int]
    name: str
    versions: _containers.RepeatedScalarFieldContainer[str]
    framework: str
    inputs: _containers.RepeatedCompositeFieldContainer[ModelMetadataResponse.TensorMetadata]
    outputs: _containers.RepeatedCompositeFieldContainer[ModelMetadataResponse.TensorMetadata]
    def __init__(self, name: _Optional[str] = ..., versions: _Optional[_Iterable[str]] = ..., framework: _Optional[str] = ..., inputs: _Optional[_Iterable[_Union[ModelMetadataResponse.TensorMetadata, _Mapping]]] = ..., outputs: _Optional[_Iterable[_Union[ModelMetadataResponse.TensorMetadata, _Mapping]]] = ...) -> None: ...

class ModelInferRequest(_message.Message):
    __slots__ = ["model_name", "model_version", "inputs", "outputs", "raw_input_contents"]
    class InferRequestedOutputTensor(_message.Message):
        __slots__ = ["name"]
        NAME_FIELD_NUMBER: _ClassVar[int]
        name: str
        def __init__(self, name: _Optional[str] = ...) -> None: ...
    MODEL_NAME_FIELD_NUMBER: _ClassVar[int]
    MODEL_VERSION_FIELD_NUMBER: _ClassVar[int]
    INPUTS_FIELD_NUMBER: _ClassVar[int]
    OUTPUTS_FIELD_NUMBER: _ClassVar[int]
    RAW_INPUT_CONTENTS_FIELD_NUMBER: _ClassVar[int]
    model_name: str
    model_version: str
    inputs: _containers.RepeatedCompositeFieldContainer[InferTensor]
    outputs: _containers.RepeatedCompositeFieldContainer[ModelInferRequest.InferRequestedOutputTensor]
    raw_input_contents: _containers.RepeatedScalarFieldContainer[bytes]
    def __init__(self, model_name: _Optional[str] = ..., model_version: _Optional[str] = ..., inputs: _Optional[_Iterable[_Union[InferTensor, _Mapping]]] = ..., outputs: _Optional[_Iterable[_Union[ModelInferRequest.InferRequestedOutputTensor, _Mapping]]] = ..., raw_input_contents: _Optional[_Iterable[bytes]] = ...) -> None: ...

class ModelInferResponse(_message.Message):
    __slots__ = ["model_name", "model_version", "outputs", "raw_output_contents"]
    MODEL_NAME_FIELD_NUMBER: _ClassVar[int]
    MODEL_VERSION_FIELD_NUMBER: _ClassVar[int]
    OUTPUTS_FIELD_NUMBER: _ClassVar[int]
    RAW_OUTPUT_CONTENTS_FIELD_NUMBER: _ClassVar[int]
    model_name: str
    model_version: str
    outputs: _containers.RepeatedCompositeFieldContainer[InferTensor]
    raw_output_contents: _containers.RepeatedScalarFieldContainer[bytes]
    def __init__(self, model_name: _Optional[str] = ..., model_version: _Optional[str] = ..., outputs: _Optional[_Iterable[_Union[InferTensor, _Mapping]]] = ..., raw_output_contents: _Optional[_Iterable[bytes]] = ...) -> None: ...



#!/usr/bin/env python

"""Package entry point."""
import pprint

from instill.configuration import global_config

if __name__ == "__main__":  # pragma: no cover
    # main()  # pylint: disable=no-value-for-parameter
    print("======================Configured Hosts======================")
    pprint.pprint(global_config.hosts)



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

import ray_pb2 as ray__pb2


class RayServiceStub(object):
    """Ray service for internal process
    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.ModelReady = channel.unary_unary(
                '/ray.serve.RayService/ModelReady',
                request_serializer=ray__pb2.ModelReadyRequest.SerializeToString,
                response_deserializer=ray__pb2.ModelReadyResponse.FromString,
                )
        self.ModelMetadata = channel.unary_unary(
                '/ray.serve.RayService/ModelMetadata',
                request_serializer=ray__pb2.ModelMetadataRequest.SerializeToString,
                response_deserializer=ray__pb2.ModelMetadataResponse.FromString,
                )
        self.ModelInfer = channel.unary_unary(
                '/ray.serve.RayService/ModelInfer',
                request_serializer=ray__pb2.ModelInferRequest.SerializeToString,
                response_deserializer=ray__pb2.ModelInferResponse.FromString,
                )


class RayServiceServicer(object):
    """Ray service for internal process
    """

    def ModelReady(self, request, context):
        """ModelReady method receives a ModelReadyRequest message and
        returns a ModelReadyResponse
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ModelMetadata(self, request, context):
        """ModelMetadata method receives a ModelMetadataRequest message and
        returns a ModelMetadataResponse
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ModelInfer(self, request, context):
        """ModelInfer method receives a ModelInferRequest message and
        returns a ModelInferResponse
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_RayServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'ModelReady': grpc.unary_unary_rpc_method_handler(
                    servicer.ModelReady,
                    request_deserializer=ray__pb2.ModelReadyRequest.FromString,
                    response_serializer=ray__pb2.ModelReadyResponse.SerializeToString,
            ),
            'ModelMetadata': grpc.unary_unary_rpc_method_handler(
                    servicer.ModelMetadata,
                    request_deserializer=ray__pb2.ModelMetadataRequest.FromString,
                    response_serializer=ray__pb2.ModelMetadataResponse.SerializeToString,
            ),
            'ModelInfer': grpc.unary_unary_rpc_method_handler(
                    servicer.ModelInfer,
                    request_deserializer=ray__pb2.ModelInferRequest.FromString,
                    response_serializer=ray__pb2.ModelInferResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'ray.serve.RayService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class RayService(object):
    """Ray service for internal process
    """

    @staticmethod
    def ModelReady(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/ray.serve.RayService/ModelReady',
            ray__pb2.ModelReadyRequest.SerializeToString,
            ray__pb2.ModelReadyResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def ModelMetadata(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/ray.serve.RayService/ModelMetadata',
            ray__pb2.ModelMetadataRequest.SerializeToString,
            ray__pb2.ModelMetadataResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def ModelInfer(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/ray.serve.RayService/ModelInfer',
            ray__pb2.ModelInferRequest.SerializeToString,
            ray__pb2.ModelInferResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



# pylint: disable=inconsistent-return-statements,no-member
import os

import grpc

from instill.utils.logger import Logger


class NotServingException(Exception):
    def __str__(self) -> str:
        return "target host is not serving"


def grpc_handler(func):
    def func_wrapper(*args, **kwargs):
        silent = kwargs.pop("silent", False)
        try:
            if not args[0].is_serving():
                raise NotServingException
            return func(*args, **kwargs)
        except grpc.RpcError as rpc_error:
            if not silent:
                Logger.w(rpc_error.code())
                Logger.w(rpc_error.details())
                os._exit(1)
        except Exception as e:
            Logger.exception(e)
            os._exit(1)

    return func_wrapper



# pylint: disable=no-member,wrong-import-position
import instill.protogen.vdp.pipeline.v1alpha.pipeline_pb2 as pipeline_pb


def create_recipe(component: list) -> pipeline_pb.Recipe:
    recipe = pipeline_pb.Recipe()
    recipe.version = "v1alpha"
    recipe.components.extend(component)

    return recipe



import instill.protogen.model.model.v1alpha.model_pb2 as model_pb
import instill.protogen.model.model.v1alpha.task_classification_pb2 as task_classification
import instill.protogen.model.model.v1alpha.task_detection_pb2 as task_detection
import instill.protogen.model.model.v1alpha.task_keypoint_pb2 as task_keypoint
import instill.protogen.model.model.v1alpha.task_ocr_pb2 as task_ocr
import instill.protogen.model.model.v1alpha.task_semantic_segmentation_pb2 as task_semantic_segmentation
import instill.protogen.model.model.v1alpha.task_text_generation_pb2 as task_text_generation
import instill.protogen.model.model.v1alpha.task_text_to_image_pb2 as task_text_to_image
import instill.protogen.vdp.pipeline.v1alpha.connector_pb2 as connector_pb
import instill.protogen.vdp.pipeline.v1alpha.pipeline_pb2 as pipeline_pb
from instill.resources.connector import Connector
from instill.resources.connector_ai import (
    InstillModelConnector,
    OpenAIConnector,
    StabilityAIConnector,
)
from instill.resources.connector_blockchain import NumbersConnector
from instill.resources.connector_data import PineconeConnector
from instill.resources.model import GithubModel, HugginfaceModel, Model
from instill.resources.operator import create_end_operator, create_start_operator
from instill.resources.pipeline import Pipeline
from instill.resources.recipe import create_recipe



from instill.clients.client import InstillClient, get_client
from instill.clients.mgmt import MgmtClient
from instill.clients.model import ModelClient
from instill.clients.pipeline import PipelineClient



# pylint: disable=no-member,wrong-import-position
from abc import ABC, abstractmethod


class Resource(ABC):
    """Base interface class for creating resources.

    Args:
        ABC (abc.ABCMeta): std abstract class
    """

    @property
    @abstractmethod
    def client(self):
        pass

    @client.setter
    @abstractmethod
    def client(self):
        pass

    @property
    @abstractmethod
    def resource(self):
        pass

    @resource.setter
    @abstractmethod
    def resource(self):
        pass



# pylint: disable=no-name-in-module
from instill.clients.mgmt import MgmtClient
from instill.clients.model import ModelClient
from instill.clients.pipeline import PipelineClient
from instill.utils.error_handler import NotServingException
from instill.utils.logger import Logger

_mgmt_client = None
_pipeline_client = None
_model_client = None
_client = None


def _get_mgmt_client() -> MgmtClient:
    global _mgmt_client

    if _mgmt_client is None:
        _mgmt_client = MgmtClient()

    return _mgmt_client


def _get_pipeline_client() -> PipelineClient:
    global _pipeline_client

    if _pipeline_client is None:
        _pipeline_client = PipelineClient(namespace=_get_mgmt_client().get_user().name)

    return _pipeline_client


def _get_model_client() -> ModelClient:
    global _model_client

    if _model_client is None:
        _model_client = ModelClient(namespace=_get_mgmt_client().get_user().name)

    return _model_client


class InstillClient:
    def __init__(self) -> None:
        self.mgmt_service = _get_mgmt_client()
        if not self.mgmt_service.is_serving():
            Logger.w("Instill Core is required")
            raise NotServingException
        self.pipeline_service = _get_pipeline_client()
        if not self.pipeline_service.is_serving():
            Logger.w("Instill VDP is not serving, VDP functionalities will not work")
        self.model_service = _get_model_client()
        if not self.model_service.is_serving():
            Logger.w(
                "Instill Model is not serving, Model functionalities will not work"
            )

    def set_instance(self, instance: str):
        self.mgmt_service.instance = instance
        self.pipeline_service.instance = instance
        self.model_service.instance = instance

    def close(self):
        if self.mgmt_service.is_serving():
            for host in self.mgmt_service.hosts.values():
                host["channel"].close()
        if self.pipeline_service.is_serving():
            for host in self.pipeline_service.hosts.values():
                host["channel"].close()
        if self.model_service.is_serving():
            for host in self.model_service.hosts.values():
                host["channel"].close()


def get_client() -> InstillClient:
    global _client

    if _client is None:
        _client = InstillClient()

    return _client



# pylint: disable=no-member,wrong-import-position,no-name-in-module
import instill.protogen.model.model.v1alpha.model_definition_pb2 as model_definition_interface
import instill.protogen.model.model.v1alpha.model_pb2 as model_interface
from instill.clients import InstillClient
from instill.resources.resource import Resource


class Model(Resource):
    def __init__(
        self,
        client: InstillClient,
        name: str,
        definition: str,
        configuration: dict,
    ) -> None:
        super().__init__()
        self.client = client
        model = client.model_service.get_model(model_name=name, silent=True)
        if model is None:
            model = client.model_service.create_model(
                name=name,
                definition=definition,
                configuration=configuration,
            )
            if model is None:
                raise BaseException("model creation failed")

        self.resource = model

    def __call__(self, task_inputs: list) -> list:
        return self.client.model_service.trigger_model(self.resource.id, task_inputs)

    @property
    def client(self):
        return self._client

    @client.setter
    def client(self, client: InstillClient):
        self._client = client

    @property
    def resource(self):
        return self._resource

    @resource.setter
    def resource(self, resource: model_interface.Model):
        self._resource = resource

    def _update(self):
        self.resource = self.client.model_service.get_model(model_name=self.resource.id)

    def get_definition(self) -> model_definition_interface.ModelDefinition:
        return self.resource.model_definition

    def get_readme(self) -> str:
        return self.client.model_service.get_model_card(self.resource.id)

    def get_state(self) -> model_interface.Model.State:
        return self.client.model_service.watch_model(self.resource.id)

    def deploy(self) -> model_interface.Model:
        self.client.model_service.deploy_model(self.resource.id)
        self._update()
        return self._resource

    def undeploy(self) -> model_interface.Model:
        self.client.model_service.undeploy_model(self.resource.id)
        self._update()
        return self._resource

    def delete(self):
        if self.resource is not None:
            self.client.model_service.delete_model(self.resource.id)


class GithubModel(Model):
    def __init__(
        self,
        client: InstillClient,
        name: str,
        model_repo: str,
        model_tag: str,
    ) -> None:
        definition = "model-definitions/github"
        configuration = {"repository": model_repo, "tag": model_tag}
        super().__init__(
            client=client,
            name=name,
            definition=definition,
            configuration=configuration,
        )


class HugginfaceModel(Model):
    def __init__(
        self,
        client: InstillClient,
        name: str,
        model_repo: str,
    ) -> None:
        configuration = {"repo_id": model_repo}
        definition = "model-definitions/huggingface"
        super().__init__(
            client=client,
            name=name,
            definition=definition,
            configuration=configuration,
        )



# pylint: disable=no-member,wrong-import-position
import time
from collections import defaultdict
from typing import Iterable, Tuple, Union

import grpc
from google.longrunning import operations_pb2
from google.protobuf import field_mask_pb2

import instill.protogen.common.healthcheck.v1alpha.healthcheck_pb2 as healthcheck
import instill.protogen.model.model.v1alpha.model_definition_pb2 as model_definition_interface

# model
import instill.protogen.model.model.v1alpha.model_pb2 as model_interface
import instill.protogen.model.model.v1alpha.model_public_service_pb2_grpc as model_service
from instill.clients import constant
from instill.clients.base import Client

# common
from instill.configuration import global_config
from instill.utils.error_handler import grpc_handler


class ModelClient(Client):
    def __init__(self, namespace: str) -> None:
        self.hosts: defaultdict = defaultdict(dict)
        self.namespace: str = namespace
        if constant.DEFAULT_INSTANCE in global_config.hosts:
            self.instance = constant.DEFAULT_INSTANCE
        elif len(global_config.hosts) == 0:
            self.instance = ""
        else:
            self.instance = list(global_config.hosts.keys())[0]

        if global_config.hosts is not None:
            for instance, config in global_config.hosts.items():
                if not config.secure:
                    channel = grpc.insecure_channel(config.url)
                    self.hosts[instance]["metadata"] = (
                        (
                            "authorization",
                            f"Bearer {config.token}",
                        ),
                    )
                else:
                    ssl_creds = grpc.ssl_channel_credentials()
                    call_creds = grpc.access_token_call_credentials(config.token)
                    creds = grpc.composite_channel_credentials(ssl_creds, call_creds)
                    channel = grpc.secure_channel(
                        target=config.url,
                        credentials=creds,
                    )
                    self.hosts[instance]["metadata"] = ""
                self.hosts[instance]["token"] = config.token
                self.hosts[instance]["channel"] = channel
                self.hosts[instance]["client"] = model_service.ModelPublicServiceStub(
                    channel
                )

    @property
    def hosts(self):
        return self._hosts

    @hosts.setter
    def hosts(self, hosts: str):
        self._hosts = hosts

    @property
    def instance(self):
        return self._instance

    @instance.setter
    def instance(self, instance: str):
        self._instance = instance

    @property
    def metadata(self):
        return self._metadata

    @metadata.setter
    def metadata(self, metadata: str):
        self._metadata = metadata

    def liveness(self) -> healthcheck.HealthCheckResponse.ServingStatus:
        resp: model_interface.LivenessResponse = self.hosts[self.instance][
            "client"
        ].Liveness(request=model_interface.LivenessRequest())
        return resp.health_check_response.status

    def readiness(self) -> healthcheck.HealthCheckResponse.ServingStatus:
        resp: model_interface.ReadinessResponse = self.hosts[self.instance][
            "client"
        ].Readiness(request=model_interface.ReadinessRequest())
        return resp.health_check_response.status

    def is_serving(self) -> bool:
        try:
            return (
                self.readiness()
                == healthcheck.HealthCheckResponse.SERVING_STATUS_SERVING
            )
        except Exception:
            return False

    @grpc_handler
    def watch_model(self, model_name: str) -> model_interface.Model.State.ValueType:
        resp: model_interface.WatchUserModelResponse = self.hosts[self.instance][
            "client"
        ].WatchUserModel(
            request=model_interface.WatchUserModelRequest(
                name=f"{self.namespace}/models/{model_name}"
            ),
            metadata=self.hosts[self.instance]["metadata"],
        )
        return resp.state

    @grpc_handler
    def create_model_local(
        self,
        model_name: str,
        model_description: str,
        model_path: str,
    ) -> model_interface.Model:
        model = model_interface.Model()
        model.id = model_name
        model.description = model_description
        model.model_definition = "model-definitions/local"

        with open(model_path, "rb") as f:
            data = f.read()
            req = model_interface.CreateUserModelBinaryFileUploadRequest(
                parent=self.namespace, model=model, content=data
            )
        create_resp: model_interface.CreateUserModelBinaryFileUploadResponse = (
            self.hosts[self.instance]["client"].CreateUserModelBinaryFileUpload(
                request_iterator=iter([req]),
                metadata=self.hosts[self.instance]["metadata"],
            )
        )

        while self.get_operation(name=create_resp.operation.name).done is not True:
            time.sleep(1)

        state = self.watch_model(model_name=model_name)
        while state == 0:
            time.sleep(1)
            state = self.watch_model(model_name=model_name)

        if state == 1:
            return self.get_model(model_name=model_name)

        raise SystemError("model creation failed")

    @grpc_handler
    def create_model(
        self,
        name: str,
        definition: str,
        configuration: dict,
    ) -> model_interface.Model:
        model = model_interface.Model()
        model.id = name
        model.model_definition = definition
        model.configuration.update(configuration)
        create_resp: model_interface.CreateUserModelResponse = self.hosts[
            self.instance
        ]["client"].CreateUserModel(
            request=model_interface.CreateUserModelRequest(
                model=model, parent=self.namespace
            ),
            metadata=self.hosts[self.instance]["metadata"],
        )

        while self.get_operation(name=create_resp.operation.name).done is not True:
            time.sleep(1)

        # TODO: due to state update delay of controller
        # TODO: should optimize this in model-backend
        time.sleep(3)

        state = self.watch_model(model_name=name)
        while state == 0:
            time.sleep(1)
            state = self.watch_model(model_name=name)

        if state == 1:
            return self.get_model(model_name=name)

        raise SystemError("model creation failed")

    @grpc_handler
    def deploy_model(self, model_name: str) -> model_interface.Model.State:
        self.hosts[self.instance]["client"].DeployUserModel(
            request=model_interface.DeployUserModelRequest(
                name=f"{self.namespace}/models/{model_name}"
            ),
            metadata=self.hosts[self.instance]["metadata"],
        )

        state = self.watch_model(model_name=model_name)
        while state not in (2, 3):
            time.sleep(1)
            state = self.watch_model(model_name=model_name)

        return state

    @grpc_handler
    def undeploy_model(self, model_name: str) -> model_interface.Model.State:
        self.hosts[self.instance]["client"].UndeployUserModel(
            request=model_interface.UndeployUserModelRequest(
                name=f"{self.namespace}/models/{model_name}"
            ),
            metadata=self.hosts[self.instance]["metadata"],
        )

        state = self.watch_model(model_name=model_name)
        while state not in (1, 3):
            time.sleep(1)
            state = self.watch_model(model_name=model_name)

        return state

    @grpc_handler
    def trigger_model(self, model_name: str, task_inputs: list) -> Iterable:
        resp: model_interface.TriggerUserModelResponse = self.hosts[self.instance][
            "client"
        ].TriggerUserModel(
            request=model_interface.TriggerUserModelRequest(
                name=f"{self.namespace}/models/{model_name}", task_inputs=task_inputs
            ),
            metadata=self.hosts[self.instance]["metadata"],
        )
        return resp.task_outputs

    @grpc_handler
    def delete_model(self, model_name: str):
        self.hosts[self.instance]["client"].DeleteUserModel(
            request=model_interface.DeleteUserModelRequest(
                name=f"{self.namespace}/models/{model_name}"
            ),
            metadata=self.hosts[self.instance]["metadata"],
        )

    @grpc_handler
    def get_model(self, model_name: str) -> model_interface.Model:
        resp: model_interface.GetUserModelResponse = self.hosts[self.instance][
            "client"
        ].GetUserModel(
            request=model_interface.GetUserModelRequest(
                name=f"{self.namespace}/models/{model_name}",
                view=model_definition_interface.VIEW_FULL,
            ),
            metadata=self.hosts[self.instance]["metadata"],
        )
        return resp.model

    @grpc_handler
    def update_model(
        self, model: model_interface.Model, mask: field_mask_pb2.FieldMask
    ) -> model_interface.Model:
        resp: model_interface.UpdateUserModelResponse = self.hosts[self.instance][
            "client"
        ].UpdateUserModel(
            request=model_interface.UpdateUserModelRequest(
                model=model,
                update_mask=mask,
            ),
            metadata=self.hosts[self.instance]["metadata"],
        )
        return resp.model

    @grpc_handler
    def lookup_model(self, model_uid: str) -> model_interface.Model:
        resp: model_interface.LookUpModelResponse = self.hosts[self.instance][
            "client"
        ].LookUpModel(
            request=model_interface.LookUpModelRequest(permalink=f"models/{model_uid}"),
            metadata=self.hosts[self.instance]["metadata"],
        )
        return resp.model

    @grpc_handler
    def get_model_card(self, model_name: str) -> model_interface.ModelCard:
        resp: model_interface.GetUserModelCardResponse = self.hosts[self.instance][
            "client"
        ].GetUserModel(
            request=model_interface.GetUserModelCardRequest(
                name=f"{self.namespace}/models/{model_name}"
            ),
            metadata=self.hosts[self.instance]["metadata"],
        )
        return resp.readme

    @grpc_handler
    def list_models(self, public=False) -> Tuple[Iterable, str, int]:
        resp: Union[
            model_interface.ListModelsResponse, model_interface.ListUserModelsResponse
        ]
        if not public:
            resp = self.hosts[self.instance]["client"].ListUserModels(
                request=model_interface.ListUserModelsRequest(parent=self.namespace),
                metadata=self.hosts[self.instance]["metadata"],
            )
        else:
            resp = self.hosts[self.instance]["client"].ListModels(
                request=model_interface.ListModelsRequest(),
                metadata=self.hosts[self.instance]["metadata"],
            )

        return resp.models, resp.next_page_token, resp.total_size

    @grpc_handler
    def get_operation(self, name: str) -> operations_pb2.Operation:
        resp: model_interface.GetModelOperationResponse = self.hosts[self.instance][
            "client"
        ].GetModelOperation(
            request=model_interface.GetModelOperationRequest(
                name=name,
            ),
            metadata=self.hosts[self.instance]["metadata"],
        )
        return resp.operation



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: ray.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\tray.proto\x12\tray.serve\"2\n\x11ModelReadyRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0f\n\x07version\x18\x02 \x01(\t\"#\n\x12ModelReadyResponse\x12\r\n\x05ready\x18\x01 \x01(\x08\"5\n\x14ModelMetadataRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0f\n\x07version\x18\x02 \x01(\t\"<\n\x0bInferTensor\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x10\n\x08\x64\x61tatype\x18\x02 \x01(\t\x12\r\n\x05shape\x18\x03 \x03(\x03\"\x8e\x02\n\x15ModelMetadataResponse\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x10\n\x08versions\x18\x02 \x03(\t\x12\x11\n\tframework\x18\x03 \x01(\t\x12?\n\x06inputs\x18\x04 \x03(\x0b\x32/.ray.serve.ModelMetadataResponse.TensorMetadata\x12@\n\x07outputs\x18\x05 \x03(\x0b\x32/.ray.serve.ModelMetadataResponse.TensorMetadata\x1a?\n\x0eTensorMetadata\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x10\n\x08\x64\x61tatype\x18\x02 \x01(\t\x12\r\n\x05shape\x18\x03 \x03(\x03\"\xf8\x01\n\x11ModelInferRequest\x12\x12\n\nmodel_name\x18\x01 \x01(\t\x12\x15\n\rmodel_version\x18\x02 \x01(\t\x12&\n\x06inputs\x18\x03 \x03(\x0b\x32\x16.ray.serve.InferTensor\x12H\n\x07outputs\x18\x06 \x03(\x0b\x32\x37.ray.serve.ModelInferRequest.InferRequestedOutputTensor\x12\x1a\n\x12raw_input_contents\x18\x07 \x03(\x0c\x1a*\n\x1aInferRequestedOutputTensor\x12\x0c\n\x04name\x18\x01 \x01(\t\"\x85\x01\n\x12ModelInferResponse\x12\x12\n\nmodel_name\x18\x01 \x01(\t\x12\x15\n\rmodel_version\x18\x02 \x01(\t\x12\'\n\x07outputs\x18\x05 \x03(\x0b\x32\x16.ray.serve.InferTensor\x12\x1b\n\x13raw_output_contents\x18\x06 \x03(\x0c\x32\xfc\x01\n\nRayService\x12K\n\nModelReady\x12\x1c.ray.serve.ModelReadyRequest\x1a\x1d.ray.serve.ModelReadyResponse\"\x00\x12T\n\rModelMetadata\x12\x1f.ray.serve.ModelMetadataRequest\x1a .ray.serve.ModelMetadataResponse\"\x00\x12K\n\nModelInfer\x12\x1c.ray.serve.ModelInferRequest\x1a\x1d.ray.serve.ModelInferResponse\"\x00\x42\x07Z\x05./rayb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'ray_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'Z\005./ray'
  _globals['_MODELREADYREQUEST']._serialized_start=24
  _globals['_MODELREADYREQUEST']._serialized_end=74
  _globals['_MODELREADYRESPONSE']._serialized_start=76
  _globals['_MODELREADYRESPONSE']._serialized_end=111
  _globals['_MODELMETADATAREQUEST']._serialized_start=113
  _globals['_MODELMETADATAREQUEST']._serialized_end=166
  _globals['_INFERTENSOR']._serialized_start=168
  _globals['_INFERTENSOR']._serialized_end=228
  _globals['_MODELMETADATARESPONSE']._serialized_start=231
  _globals['_MODELMETADATARESPONSE']._serialized_end=501
  _globals['_MODELMETADATARESPONSE_TENSORMETADATA']._serialized_start=438
  _globals['_MODELMETADATARESPONSE_TENSORMETADATA']._serialized_end=501
  _globals['_MODELINFERREQUEST']._serialized_start=504
  _globals['_MODELINFERREQUEST']._serialized_end=752
  _globals['_MODELINFERREQUEST_INFERREQUESTEDOUTPUTTENSOR']._serialized_start=710
  _globals['_MODELINFERREQUEST_INFERREQUESTEDOUTPUTTENSOR']._serialized_end=752
  _globals['_MODELINFERRESPONSE']._serialized_start=755
  _globals['_MODELINFERRESPONSE']._serialized_end=888
  _globals['_RAYSERVICE']._serialized_start=891
  _globals['_RAYSERVICE']._serialized_end=1143
# @@protoc_insertion_point(module_scope)



import json
import os
import typing as t
from pathlib import Path

import yaml
from pydantic import BaseModel

CONFIG_DIR = Path(
    os.getenv(
        "INSTILL_SYSTEM_CONFIG_PATH",
        Path.home() / ".config/instill/sdk/python/",
    )
)


class _InstillHost(BaseModel):
    url: str
    secure: bool
    token: str


class _Config(BaseModel):
    hosts: t.Dict[str, _InstillHost] = {}


class Configuration:
    def __init__(self) -> None:
        self._config: _Config

        CONFIG_DIR.mkdir(exist_ok=True)

    @property
    def hosts(self) -> t.Dict[str, _InstillHost]:
        return self._config.hosts

    def load(self) -> None:
        path = CONFIG_DIR / "config.yml"
        if not path.exists():
            self._config = _Config()
            return
        try:
            with open(path, "r", encoding="utf-8") as c:
                self._config = _Config.model_validate(
                    yaml.load(c, Loader=yaml.FullLoader)
                )
        except Exception as e:
            raise BaseException(f"Invalid configuration file at '{path}'") from e

    def save(self) -> None:
        path = CONFIG_DIR / "config.yml"

        CONFIG_DIR.mkdir(exist_ok=True)

        with open(path, "w", encoding="utf-8") as c:
            yaml.dump(
                json.loads(
                    self._config.model_dump_json(
                        exclude_none=True,
                    )
                ),
                c,
            )

    def set_default(self, url: str, token: str, secure: bool):
        self._config.hosts["default"] = _InstillHost(
            url=url, secure=secure, token=token
        )


global_config = Configuration()
global_config.load()
